#ifndef __FBLIB_H__
#define __FBLIB_H__

#include <fcntl.h>
#include <linux/fb.h>
#include <ws/misc.h>

#ifndef DEFINED_STRUCT_FONTINFO_TAG
#define DEFINED_STRUCT_FONTINFO_TAG 1
struct fontinfo_tag
 {
  signed char *data;
  int table[96];
  int yofs,fh,oy;
 };
#endif
typedef struct fontinfo_tag *Font; 

typedef UnsignedChar ColourValue;

typedef struct
 {
  ColourValue blue;
  ColourValue green;
  ColourValue red;
  ColourValue transp;
 } Colour;

typedef Int Coordinate;
typedef UnsignedInt Size;

struct FrameBufferT;
struct FrameBufferT
 {
  char *fbp;
  struct fb_var_screeninfo vinfo;
  struct fb_fix_screeninfo finfo;
  long int screensize;
  
  char *(*pixelLocation)(struct FrameBufferT *fb, Coordinate x, Coordinate y);
  void (*putPixelLocation)(char *location, Colour colour);
  void (*putPixel)(struct FrameBufferT *fb, Coordinate x, Coordinate y, Colour colour);
  void (*drawVLine)(struct FrameBufferT *fb, Coordinate x, Coordinate y, Size height, Colour colour);
  void (*drawHLine)(struct FrameBufferT *fb, Coordinate x, Coordinate y, Size width, Colour colour);
  void (*drawRectangle)(struct FrameBufferT *fb, Coordinate x, Coordinate y, Size dx, Size dy, Colour colour);
  void (*drawFilledRectangle)(struct FrameBufferT *fb, Coordinate x, Coordinate y, Size dx, Size dy, Colour colour);
  void (*draw3DRectangle)(struct FrameBufferT *fb, Coordinate x, Coordinate y, Size dx, Size dy, Colour colour1, Colour colour2);
  void (*copyArea)(struct FrameBufferT *tfb, struct FrameBufferT *ffb, Coordinate tx, Coordinate ty, Coordinate fx, Coordinate fy, Size w, Size h);
  void (*moveArea)(struct FrameBufferT *tfb, struct FrameBufferT *ffb, Coordinate tx, Coordinate ty, Coordinate fx, Coordinate fy, Size w, Size h);
  void (*drawOverlay)(struct FrameBufferT *fb, struct FrameBufferT *overlay, Coordinate x, Coordinate y, Coordinate ox, Coordinate oy, Size w, Size h);
  size_t (*drawText)(struct FrameBufferT *fb, Coordinate x, Coordinate y, Size w, Size h, Coordinate vx, Coordinate vy, Size vw, Size vh, char *text, Font font, Colour colour);
 };
typedef struct FrameBufferT FrameBuffer;

#define MACcreatPixelFormat3(MACsname, MACcname, MAC0fname, MAC1fname, MAC2fname, MAC0cname, MAC1cname, MAC2cname) \
typedef struct \
 { \
  uint MAC0fname : MACcname##_WIDTH_##MAC0cname; \
  uint MAC1fname : MACcname##_WIDTH_##MAC1cname; \
  uint MAC2fname : MACcname##_WIDTH_##MAC2cname; \
 } __attribute__((packed)) MACsname;

#define MACcreatPixelFormat4(MACsname, MACcname, MAC0fname, MAC1fname, MAC2fname, MAC3fname, MAC0cname, MAC1cname, MAC2cname, MAC3cname) \
typedef struct \
 { \
  uint MAC0fname : MACcname##_WIDTH_##MAC0cname; \
  uint MAC1fname : MACcname##_WIDTH_##MAC1cname; \
  uint MAC2fname : MACcname##_WIDTH_##MAC2cname; \
  uint MAC3fname : MACcname##_WIDTH_##MAC3cname; \
 } __attribute__((packed)) MACsname;

#define PIXEL8_WIDTH_BLUE 3
#define PIXEL8_WIDTH_GREEN 2
#define PIXEL8_WIDTH_RED 3
#define PIXEL8_WIDTH_TRANSP 0
#define PIXEL8_OFFSET_BLUE 0
#define PIXEL8_OFFSET_GREEN 3
#define PIXEL8_OFFSET_RED 5
#define PIXEL8_OFFSET_TRANSP 8
MACcreatPixelFormat3(Pixel8, PIXEL8, blue, green, red, BLUE, GREEN, RED)

#define PIXEL16_WIDTH_BLUE 5
#define PIXEL16_WIDTH_GREEN 6
#define PIXEL16_WIDTH_RED 5
#define PIXEL16_WIDTH_TRANSP 0
#define PIXEL16_OFFSET_BLUE 0
#define PIXEL16_OFFSET_GREEN 5
#define PIXEL16_OFFSET_RED 11
#define PIXEL16_OFFSET_TRANSP 16
MACcreatPixelFormat3(Pixel16, PIXEL16, blue, green, red, BLUE, GREEN, RED)

#define PIXEL24_WIDTH_BLUE 8
#define PIXEL24_WIDTH_GREEN 8
#define PIXEL24_WIDTH_RED 8
#define PIXEL24_WIDTH_TRANSP 0
#define PIXEL24_OFFSET_BLUE 0
#define PIXEL24_OFFSET_GREEN 8
#define PIXEL24_OFFSET_RED 16
#define PIXEL24_OFFSET_TRANSP 24
MACcreatPixelFormat3(Pixel24, PIXEL24, blue, green, red, BLUE, GREEN, RED)

#define PIXEL32_WIDTH_BLUE 8
#define PIXEL32_WIDTH_GREEN 8
#define PIXEL32_WIDTH_RED 8
#define PIXEL32_WIDTH_TRANSP 8
#define PIXEL32_OFFSET_BLUE 0
#define PIXEL32_OFFSET_GREEN 8
#define PIXEL32_OFFSET_RED 16
#define PIXEL32_OFFSET_TRANSP 24
MACcreatPixelFormat4(Pixel32, PIXEL32, blue, green, red, transp, BLUE, GREEN, RED, TRANSP)


#define MACsetPixelRedPixel8(MAClocation, MACred)       ((Pixel8 *) (MAClocation))->red    = MACred>>(8 - PIXEL8_WIDTH_RED)
#define MACsetPixelGreenPixel8(MAClocation, MACgreen)   ((Pixel8 *) (MAClocation))->green  = MACgreen>>(8 - PIXEL8_WIDTH_GREEN)
#define MACsetPixelBluePixel8(MAClocation, MACblue)     ((Pixel8 *) (MAClocation))->blue   = MACblue>>(8 - PIXEL8_WIDTH_BLUE)
#define MACsetPixelTranspPixel8(MAClocation, MACtransp)  ({MACtransp;})

#define MACsetPixelRedPixel16(MAClocation, MACred)       ((Pixel16 *) (MAClocation))->red    = MACred>>(8 - PIXEL16_WIDTH_RED)
#define MACsetPixelGreenPixel16(MAClocation, MACgreen)   ((Pixel16 *) (MAClocation))->green  = MACgreen>>(8 - PIXEL16_WIDTH_GREEN)
#define MACsetPixelBluePixel16(MAClocation, MACblue)     ((Pixel16 *) (MAClocation))->blue   = MACblue>>(8 - PIXEL16_WIDTH_BLUE)
#define MACsetPixelTranspPixel16(MAClocation, MACtransp)  ({MACtransp;})

#define MACsetPixelRedPixel24(MAClocation, MACred)       ((Pixel24 *) (MAClocation))->red    = MACred>>(8 - PIXEL24_WIDTH_RED)
#define MACsetPixelGreenPixel24(MAClocation, MACgreen)   ((Pixel24 *) (MAClocation))->green  = MACgreen>>(8 - PIXEL24_WIDTH_GREEN)
#define MACsetPixelBluePixel24(MAClocation, MACblue)     ((Pixel24 *) (MAClocation))->blue   = MACblue>>(8 - PIXEL24_WIDTH_BLUE)
#define MACsetPixelTranspPixel24(MAClocation, MACtransp) ({MACtransp;})

#define MACsetPixelRedPixel32(MAClocation, MACred)       ((Pixel32 *) (MAClocation))->red    = MACred>>(8 - PIXEL32_WIDTH_RED)
#define MACsetPixelGreenPixel32(MAClocation, MACgreen)   ((Pixel32 *) (MAClocation))->green  = MACgreen>>(8 - PIXEL32_WIDTH_GREEN)
#define MACsetPixelBluePixel32(MAClocation, MACblue)     ((Pixel32 *) (MAClocation))->blue   = MACblue>>(8 - PIXEL32_WIDTH_BLUE)
#define MACsetPixelTranspPixel32(MAClocation, MACtransp) ((Pixel32 *) (MAClocation))->transp = MACtransp>>(8 - PIXEL32_WIDTH_TRANSP)


#define MACgetPixelRedPixel8(MAClocation)                ASL(((ColourValue) ((Pixel8 *) (MAClocation))->red), (8 - PIXEL8_WIDTH_RED))
#define MACgetPixelGreenPixel8(MAClocation)              ASL(((ColourValue) ((Pixel8 *) (MAClocation))->green), (8 - PIXEL8_WIDTH_GREEN))
#define MACgetPixelBluePixel8(MAClocation)               ASL(((ColourValue) ((Pixel8 *) (MAClocation))->blue), (8 - PIXEL8_WIDTH_BLUE))
#define MACgetPixelTranspPixel8(MAClocation)             0

#define MACgetPixelRedPixel16(MAClocation)               ASL(((ColourValue) ((Pixel16 *) (MAClocation))->red), (8 - PIXEL16_WIDTH_RED))
#define MACgetPixelGreenPixel16(MAClocation)             ASL(((ColourValue) ((Pixel16 *) (MAClocation))->green), (8 - PIXEL16_WIDTH_GREEN))
#define MACgetPixelBluePixel16(MAClocation)              ASL(((ColourValue) ((Pixel16 *) (MAClocation))->blue), (8 - PIXEL16_WIDTH_BLUE))
#define MACgetPixelTranspPixel16(MAClocation)            0

#define MACgetPixelRedPixel24(MAClocation)               ASL(((ColourValue) ((Pixel24 *) (MAClocation))->red), (8 - PIXEL24_WIDTH_RED))
#define MACgetPixelGreenPixel24(MAClocation)             ASL(((ColourValue) ((Pixel24 *) (MAClocation))->green), (8 - PIXEL24_WIDTH_GREEN))
#define MACgetPixelBluePixel24(MAClocation)              ASL(((ColourValue) ((Pixel24 *) (MAClocation))->blue), (8 - PIXEL24_WIDTH_BLUE))
#define MACgetPixelTranspPixel24(MAClocation)            0

#define MACgetPixelRedPixel32(MAClocation)               ASL(((ColourValue) ((Pixel32 *) (MAClocation))->red), (8 - PIXEL32_WIDTH_RED))
#define MACgetPixelGreenPixel32(MAClocation)             ASL(((ColourValue) ((Pixel32 *) (MAClocation))->green), (8 - PIXEL32_WIDTH_GREEN))
#define MACgetPixelBluePixel32(MAClocation)              ASL(((ColourValue) ((Pixel32 *) (MAClocation))->blue), (8 - PIXEL32_WIDTH_BLUE))
#define MACgetPixelTranspPixel32(MAClocation)            ASL(((ColourValue) ((Pixel32 *) (MAClocation))->transp), (8 - PIXEL32_WIDTH_TRANSP))


#define MACmakeAllFormats(MACname) \
MAC##MACname##Creat(Pixel8) \
MAC##MACname##Creat(Pixel16) \
MAC##MACname##Creat(Pixel24) \
MAC##MACname##Creat(Pixel32)
#define MACmakeAllDepths(MACname) \
MAC##MACname##Creat(8) \
MAC##MACname##Creat(16) \
MAC##MACname##Creat(24) \
MAC##MACname##Creat(32)
#define MACformatFromDepth(MACdepth) Pixel##MACdepth


#define MACdrawPixelLocation(MACpixelformat, MAClocation, red, green, blue, transp) \
 do { \
  MACsetPixelRed##MACpixelformat(MAClocation, \
                                 (ColourValue) ((  ((unsigned long int) MACgetPixelRed##MACpixelformat(MAClocation) * (unsigned long int) transp) \
                                                 + (unsigned long int) red * (unsigned long int) (255 - transp)) / 255)); \
  MACsetPixelGreen##MACpixelformat(MAClocation, \
                                 (ColourValue) ((  ((unsigned long int) MACgetPixelGreen##MACpixelformat(MAClocation) * (unsigned long int) transp) \
                                                 + (unsigned long int) green * (unsigned long int) (255 - transp)) / 255)); \
  MACsetPixelBlue##MACpixelformat(MAClocation, \
                                 (ColourValue) ((  ((unsigned long int) MACgetPixelBlue##MACpixelformat(MAClocation) * (unsigned long int) transp) \
                                                 + (unsigned long int) blue * (unsigned long int) (255 - transp)) / 255)); \
 } while(0)


#define MACpixelLocation(MACfbp, MACvinfo, MACfinfo, MACx, MACy) ((MACfbp) + ((MACx) + (MACvinfo).xoffset) * ((MACvinfo).bits_per_pixel / 8) + ((MACy) + (MACvinfo).yoffset) * (MACfinfo).line_length)

#define MACputPixelLocation(MACpixelformat, MAClocation, MACred, MACgreen, MACblue, MACtransp) \
 do { \
  MACsetPixelRed##MACpixelformat(MAClocation, MACred); \
  MACsetPixelGreen##MACpixelformat(MAClocation, MACgreen); \
  MACsetPixelBlue##MACpixelformat(MAClocation, MACblue); \
  MACsetPixelTransp##MACpixelformat(MAClocation, MACtransp); \
 } while (0)

#define MACputPixel(MACpixelformat, MACfbp, MACvinfo, MACfinfo, MACx, MACy, MACred, MACgreen, MACblue, MACtransp) \
  MACputPixelLocation(MACpixelformat, MACpixelLocation(MACfbp, MACvinfo, MACfinfo, MACx, MACy), MACred, MACgreen, MACblue, MACtransp)

#define MACgetPixelLocation(MACpixelformat, MAClocation, MACred, MACgreen, MACblue, MACtransp) \
 do { \
  MACred = MACgetPixelRed##MACpixelformat(MAClocation); \
  MACgreen = MACgetPixelGreen##MACpixelformat(MAClocation); \
  MACblue = MACgetPixelBlue##MACpixelformat(MAClocation); \
  MACtransp = MACgetPixelTransp##MACpixelformat(MAClocation); \
 } while (0)

#define MACgetPixel(MACpixelformat, MACfbp, MACvinfo, MACfinfo, MACx, MACy, MACred, MACgreen, MACblue, MACtransp) \
  MACgetPixelLocation(MACpixelformat, MACpixelLocation(MACfbp, MACvinfo, MACfinfo, MACx, MACy), MACred, MACgreen, MACblue, MACtransp)

#define offscreenBufferNew(MACob, MACpixelformat, MACw, MACh) offscreenBufferNew##MACpixelformat(MACob, MACw, MACh)
#define offscreenBufferNewPixel8(MACob, MACw, MACh) offscreenBufferNewFromDepth(MACob, 8, MACw, MACh)
#define offscreenBufferNewPixel16(MACob, MACw, MACh) offscreenBufferNewFromDepth(MACob, 16, MACw, MACh)
#define offscreenBufferNewPixel24(MACob, MACw, MACh) offscreenBufferNewFromDepth(MACob, 24, MACw, MACh)
#define offscreenBufferNewPixel32(MACob, MACw, MACh) offscreenBufferNewFromDepth(MACob, 32, MACw, MACh)

extern int  offscreenBufferNewFromDepth(FrameBuffer *ob, Int depth, Size w, Size h);
extern int  offscreenBufferNewFromExisting(FrameBuffer *ob, FrameBuffer *fb, Size w, Size h);
extern void offscreenBufferDestroy(FrameBuffer *ob);
extern int  frameBufferNew(FrameBuffer *fb, int fbfd);
extern void frameBufferDestroy(FrameBuffer *fb);

extern Size drawTextHeight(Size w, char *text, Font font);


extern int  frameBufferInit(FrameBuffer *fb, Boolean active);
extern void offscreenBufferSetup(FrameBuffer *ob, Size w, Size h, char *name);
extern int  offscreenBufferSetupFull(FrameBuffer *ob, Int depth, Size w, Size h, char *name);
extern int  offscreenBufferMalloc(FrameBuffer *ob);


#endif /* __FBLIB_H__ */
