#ifndef __MOUSELIB_H__
#define __MOUSELIB_H__

#include <ws/fblib.h>


/* OK, so most mices sends 3 or 5 bytes. So this should be guaranteed
 * to be enought :) */
#define MOUSE_PACKAGE_MAX_SIZE 10
#define MOUSE_DIMENSIONS 2

typedef UnsignedInt MouseButtons;

struct MouseT
 {
  int           fd;
  UnsignedChar  package[MOUSE_PACKAGE_MAX_SIZE];
  size_t        package_size;
  struct
   {
    unsigned char mask;
    unsigned char shadow;
   }            fingerprint[2];
  size_t        read;
  void        (*reader)(struct MouseT *mouse);
  SignedInt     direction[MOUSE_DIMENSIONS];
  MouseButtons  buttons;
  Coordinate    position[MOUSE_DIMENSIONS];
  Size          size[MOUSE_DIMENSIONS];
  void        (*handler)(Coordinate position[MOUSE_DIMENSIONS], MouseButtons buttons, void *data);
  void         *handlerData;
  void        (*setup)(struct MouseT *mouse);
 };
typedef struct MouseT Mouse;

typedef void (*MouseHandler)(Coordinate position[MOUSE_DIMENSIONS], MouseButtons buttons, void *data);
typedef void (*MouseInit)(Mouse *mouse);
typedef void (*MouseSetup)(Mouse *mouse);

extern void      mouseCreat      (Mouse *mouse, int fd, MouseInit hw, MouseHandler handler, void *handlerData, Size size[MOUSE_DIMENSIONS], Coordinate position[MOUSE_DIMENSIONS]);
extern void      mouseDestroy    (Mouse *mouse);
extern int       mouseRead       (Mouse *mouse);
extern void      mouseSetHandler (Mouse *mouse, MouseHandler handler, void *handlerData);
extern void      mouseSetPosition(Mouse *mouse, Coordinate position[MOUSE_DIMENSIONS]);
extern MouseInit mouseFromName   (String name);

#endif /* __MOUSELIB_H__ */
