#ifndef __WSLIB_H__
#define __WSLIB_H__

#include <sys/types.h>
#include <ws/rhstack.h>
#include <ws/fblib.h>
#include <ws/mdispatchlib.h>
#include <ws/kbdispatchlib.h>
#include <ws/wslib.config.h>


/*
#define CONFIG_RUNTIME_CASTCHECK
*/


typedef Coordinate WidgetSize;
typedef Coordinate WidgetOffset;
#define MAX_INFINITE 0
#define MIN_UNEXISTENT 0

/* These are versions of MIN and MAX that can handle
 * MAX_INFINITE/MIN_UNEXISTENT */
/* These two are for max values */
#define MMIN(MACx, MACy) ((MACx) == MAX_INFINITE ? \
			  (MACy) : \
			  ((MACy) == MAX_INFINITE ? \
			   (MACx) : \
			   MIN((MACx), (MACy))))
#define MMAX(MACx, MACy) ((MACx) == MAX_INFINITE ? \
			  (MACx) : \
			  ((MACy) == MAX_INFINITE ? \
			   (MACy) : \
			   MAX((MACx), (MACy))))
/* These two are for MIN values */
#define UMIN(MACx, MACy) MIN(MACx, MACy)
#define UMAX(MACx, MACy) MAX(MACx, MACy)


enum WidgetNumbers
 {
  WN_Widget,
  WN_Container,
  WN_Bin,
  WN_Frame,
  WN_Box,
  WN_Rect,
  WN_Text,
  WN_Image,
  WN_Sensor,
  WN_Input
 };

typedef enum
 {
  /* Widget types */
  #define WN2WT(name) WT_##name = 1 << WN_##name
  WN2WT(Widget),
  WN2WT(Container),
  WN2WT(Bin),
  WN2WT(Frame),
  WN2WT(Box),
  WN2WT(Rect),
  WN2WT(Text),
  WN2WT(Image),
  WN2WT(Sensor),
  WN2WT(Input),

  /* Widget inheritances */
  WTI_Widget    = WT_Widget,
  WTI_Container = WT_Container | WTI_Widget,
  WTI_Bin       = WT_Bin       | WTI_Container,
  WTI_Frame     = WT_Frame     | WTI_Bin,
  WTI_Box       = WT_Box       | WTI_Container,
  WTI_Rect      = WT_Rect      | WTI_Widget,
  WTI_Text      = WT_Text      | WTI_Widget,
  WTI_Image     = WT_Image     | WTI_Widget,
  WTI_Sensor    = WT_Sensor    | WTI_Bin,
  WTI_Input     = WT_Input     | WTI_Bin
 } WidgetType;

#define __CONVERT__(MACtype, MACobj) (*((MACtype *) &(MACobj)))
#define __PCONVERT__(MACtype, MACpoint) (*((MACtype *) MACpoint))
#define __CONVERTP__(MACtype, MACobj) ((MACtype *) &(MACobj))
#define __PCONVERTP__(MACtype, MACpoint) ((MACtype *) MACpoint)

#define ISOFTYPE(MACtype, MACpoint) (__PCONVERT__(Widget, MACpoint).type & WT_##MACtype)

#ifdef CONFIG_RUNTIME_CASTCHECK
#define CONVERT(MACtype, MACobj) ((ISOFTYPE(MACtype, &MACobj) ? 0 : castError("CONVERT", __FILE__, __LINE__, __CONVERT__(Widget, MACobj).type, WT_##MACtype##)), __CONVERT__(MACtype, MACobj))
#define PCONVERT(MACtype, MACpoint) ((ISOFTYPE(MACtype, MACpoint) ? 0 : castError("PCONVERT", __FILE__, __LINE__, __PCONVERT__(Widget, MACpoint).type, WT_##MACtype##)), __PCONVERT__(MACtype, MACpoint))
#define CONVERTP(MACtype, MACobj) ((ISOFTYPE(MACtype, &MACobj) ? 0 : castError("CONVERTP", __FILE__, __LINE__, __CONVERT__(Widget, MACobj).type, WT_##MACtype##)), __CONVERTP__(MACtype, MACobj))
#define PCONVERTP(MACtype, MACpoint) ((ISOFTYPE(MACtype, MACpoint) ? 0 : castError("PCONVERTP", __FILE__, __LINE__, __PCONVERT__(Widget, MACpoint).type, WT_##MACtype##)), __PCONVERTP__(MACtype, MACpoint))
#else
#define CONVERT __CONVERT__
#define PCONVERT __PCONVERT__
#define CONVERTP __CONVERTP__
#define PCONVERTP __PCONVERTP__
#endif

struct WidgetT;
typedef struct WidgetT *WidgetPointer;
#define WidgetPointerCopy(x) x
#define WidgetPointerDelete(x)
struct WidgetT
 {
  WidgetType        type;
  WidgetPointer     parent;
  /* Requested sizes */
  WidgetSize        minWidth;
  WidgetSize        minHeight;
  WidgetSize        maxWidth;
  WidgetSize        maxHeight;
  /* Allocated sizes */
  WidgetSize        width;
  WidgetSize        height;
  /* Position of window to scroll inside */
  WidgetSize        windowX;
  WidgetSize        windowY;
  /* Size of window to scroll inside */
  WidgetSize        windowWidth;
  WidgetSize        windowHeight;
  /* Position of the currently visible area of the window (relative
   * to the left/top of the window) */
  WidgetSize        visibleX;
  WidgetSize        visibleY;
  /* Size of the currently visible area of the window */
  WidgetSize        visibleWidth;
  WidgetSize        visibleHeight;
  /* Scroll position */
  WidgetSize        scrollX;
  WidgetSize        scrollY;

  /* Old position used when moving widgets */
  /* Position of window to scroll inside */
  WidgetSize        oldWindowX;
  WidgetSize        oldWindowY;
  /* Size of window to scroll inside */
  WidgetSize        oldWindowWidth;
  WidgetSize        oldWindowHeight;
  /* Position of the currently visible area of the window (relative
   * to the left/top of the window) */
  WidgetSize        oldVisibleX;
  WidgetSize        oldVisibleY;
  /* Size of the currently visible area of the window */
  WidgetSize        oldVisibleWidth;
  WidgetSize        oldVisibleHeight;
  /* Scroll position */
  WidgetSize        oldScrollX;
  WidgetSize        oldScrollY;

  /* Let the widget calculate the minimum and maximum space it wants */
  void (*calculateRequestedWidth) (WidgetPointer widget);
  void (*calculateRequestedHeight)(WidgetPointer widget);
  /* Let the widget calculate the size of its graphical elements
   * while it knows its own size */
  void (*calculateWidth)          (WidgetPointer widget);
  void (*calculateHeight)         (WidgetPointer widget);
  void (*show)                    (FrameBuffer *fb, WidgetPointer widget);
  void (*move)                    (FrameBuffer *fb, WidgetPointer widget);
  void (*hide)                    (WidgetPointer widget);
  void (*scroll)                  (WidgetPointer widget, WidgetSize x, WidgetSize y);
 };
typedef struct WidgetT Widget;
extern void widgetCreat(WidgetPointer this);
STACKDECL(WidgetPointer)

struct ContainerT;
typedef struct ContainerT *ContainerPointer;
struct ContainerT
 {
  Widget  parent;
  Boolean scrollingChildren;
  Colour  bg;
  /* Set this if you have made changes to the number of children or
   * whatso ever to force a recalculation of X positions and
   * visibilities */
  Boolean childChanges;
  /* Let the container calculate where to show its graphical
   * elements, now when it knows its own size and position */
  void (*calculateXPositions)   (ContainerPointer widget);
  void (*calculateYPositions)   (ContainerPointer widget);
  /* Let the container calculate which of its graphical elements, and
   * how much of them, to show, now when it knows what part of it is
   * visible and the positions of all its elements */
  void (*calculateVisibleWidth) (ContainerPointer widget);
  void (*calculateVisibleHeight)(ContainerPointer widget);
  void (*calculateChildVisibleWidth) (ContainerPointer widget, WidgetPointer child);
  void (*calculateChildVisibleHeight)(ContainerPointer widget, WidgetPointer child);
 };
typedef struct ContainerT Container;
extern void containerCreat(ContainerPointer this, Boolean scrollingChildren, Colour bg);

struct BinT;
typedef struct BinT *BinPointer;
struct BinT
 {
  Container          parent;
  WidgetPointer      child;
 };
typedef struct BinT Bin;
extern void binCreat(BinPointer this, WidgetPointer child, Boolean scrollingChildren, Colour bc);

struct FrameT;
typedef struct FrameT *FramePointer;
struct FrameT
 {
  Bin                parent;
  WidgetSize         leftMargin;
  WidgetSize         rightMargin;
  WidgetSize         topMargin;
  WidgetSize         bottomMargin;
  Boolean            fillHorizontal;
  Boolean            fillVertical;
 };
typedef struct FrameT Frame;
extern void frameCreat(FramePointer this, WidgetPointer child, Boolean scrollingChildren, WidgetSize leftMargin, WidgetSize rightMargin, WidgetSize topMargin, WidgetSize bottomMargin, Boolean fillHorizontal, Boolean fillVertical, Colour bc);

struct BoxT;
typedef struct BoxT *BoxPointer;
struct BoxT
 {
  Container          parent;
  Boolean            horizontal;
  WidgetPointerStack children;
  void (*pushChild)(BoxPointer widget, WidgetPointer child);
  void (*popChild)(BoxPointer widget);
 };
typedef struct BoxT Box;
extern void boxCreat(BoxPointer this, Boolean horizontal, Boolean scrollingChildren, Colour bg);

struct RectT;
typedef struct RectT *RectPointer;
struct RectT
 {
  Widget             parent;
  Colour             colour;
 };
typedef struct RectT Rect;
extern void rectCreat(RectPointer this, Colour colour, WidgetSize minWidth, WidgetSize maxWidth, WidgetSize minHeight, WidgetSize maxHeight);

struct TextT;
typedef struct TextT *TextPointer;
struct TextT
 {
  Widget             parent;
  String             text;
  Font               font;
  Colour             fg;
  Colour             bg;
  void               (*setText)(TextPointer widget, String text);
#ifdef TEXT_BEAUTIFULLSCROLL
#ifdef TEXT_BACKINGSTORE
  FrameBuffer        backingStore;
#else
  Boolean            textChanged;
#endif
#endif
 };
typedef struct TextT Text;
extern void textCreat(TextPointer this, String text, Font font, Colour fg, Colour bg, WidgetSize minWidth, WidgetSize maxWidth);

struct ImageT;
typedef struct ImageT *ImagePointer;
struct ImageT
 {
  Widget             parent;
  FrameBuffer       *image;
 };
typedef struct ImageT Image;
extern void imageCreat(ImagePointer this, FrameBuffer *image);

struct SensorT;
typedef struct SensorT *SensorPointer;
struct SensorT
 {
  Bin                parent;
  MouseButtons       buttons;
  MouseButtons       mask;
  MDispatcher       *dispatcher;
  MDispatcherEntry  *dispatchEntry;
  void             (*handler)(struct SensorT *sensor, Coordinate x, Coordinate y, MouseButtons buttons, void* data);
  void              *handlerData;
 };
typedef struct SensorT Sensor;
typedef void (*SensorHandler)(SensorPointer sensor, Coordinate x, Coordinate y, MouseButtons buttons, void* data);
extern void sensorCreat(SensorPointer this, MDispatcher *dispatcher, SensorHandler handler, MouseButtons buttons, MouseButtons mask, void *handlerData, WidgetPointer child, Boolean scrollingChildren, Colour bc);

struct InputT;
typedef struct InputT *InputPointer;
struct InputT
 {
  Bin                  parent;
  KBDispatcher        *dispatcher;
  long                 dispatchEntry;
  void               (*handler)(struct InputT *input, Key key, void* data);
  void               (*activate)(struct InputT *input, void* data);
  void               (*deactivate)(struct InputT *input, void* data);
  void                *handlerData;
 };
typedef struct InputT Input;
typedef void (*InputHandler)(InputPointer input, Key key, void* data);
typedef void (*InputActivateHandler)(InputPointer input, void* data);
extern void inputCreat(InputPointer this, KBDispatcher *dispatcher, InputHandler handler, InputActivateHandler activate, InputActivateHandler deactivate, void *handlerData, WidgetPointer child, Boolean scrollingChildren, Colour bc);

extern void castError(char *conversion, char *filename, int linenr, WidgetType fromType, WidgetType toType);
extern void topplevelCalculateXPosition    (FrameBuffer *fb, WidgetPointer widget);
extern void topplevelCalculateYPosition    (FrameBuffer *fb, WidgetPointer widget);
extern void topplevelCalculateWidth        (FrameBuffer *fb, WidgetPointer widget);
extern void topplevelCalculateHeight       (FrameBuffer *fb, WidgetPointer widget);
extern void topplevelCalculateVisibleWidth (FrameBuffer *fb, WidgetPointer widget);
extern void topplevelCalculateVisibleHeight(FrameBuffer *fb, WidgetPointer widget);
extern void topplevelShow                  (FrameBuffer *fb, WidgetPointer widget);

#endif /* __WSLIB_H__ */
